@@ -9,7 +9,7 @@ from glob import glob
 import html
 from collections import namedtuple, OrderedDict
 
-__version__ = '0.13.0'
+__version__ = '0.15.0'
 __author__ = 'Danny Lin'
 __author_email__ = 'danny0838@gmail.com'
 __homepage__ = 'https://github.com/danny0838/sts-lib'
@@ -64,6 +64,7 @@ class Unicode():
         i = pos
         total = len(text)
         length = 1
+        is_ids = False
 
         while length and i < total:
             code = ord(text[i])
@@ -71,13 +72,35 @@ class Unicode():
             # check if the current char is a prefix composer
             if code == 0x303E:
                 # ideographic variation indicator
+                is_ids = True
                 length += 1
             elif 0x2FF0 <= code <= 0x2FF1 or 0x2FF4 <= code <= 0x2FFB:
                 # IDS binary operator
+                is_ids = True
                 length += 2
             elif 0x2FF2 <= code <= 0x2FF3:
                 # IDS trinary operator
+                is_ids = True
                 length += 3
+            elif is_ids and not (0x4E00 <= code <= 0x9FFF  # CJK unified
+                    or 0x3400 <= code <= 0x4DBF or 0x20000 <= code <= 0x3FFFF  # Ext-A, ExtB+
+                    or 0xF900 <= code <= 0xFAFF or 0x2F800 <= code <= 0x2FA1F  # Compatibility
+                    or 0x2E80 <= code <= 0x2FDF  # Radical
+                    or 0x31C0 <= code <= 0x31EF  # Stroke
+                    or 0xE000 <= code <= 0xF8FF or 0xF0000 <= code <= 0x1FFFFF  # Private
+                    or code == 0xFF1F  # ？
+                    or 0xFE00 <= code <= 0xFE0F or 0xE0100 <= code <= 0xE01EF  # VS
+                    ):
+                # check for a valid IDS to avoid a breaking on e.g.:
+                #
+                #     IDS包括⿰⿱⿲⿳⿴⿵⿶⿷⿸⿹⿺⿻，可用於…
+                #
+                # - IDS := Ideographic | Radical | CJK_Stroke | Private Use
+                #        | U+FF1F | IDS_BinaryOperator IDS IDS
+                #        | IDS_TrinaryOperator IDS IDS IDS
+                #
+                # - We also allow IVI and VS in an IDS.
+                break
 
             # check if the next char is a postfix composer
             if i + 1 < total:
@@ -85,7 +108,7 @@ class Unicode():
                 if 0xFE00 <= code <= 0xFE0F:
                     # variation selectors
                     length += 1
-                elif 0xE0100 <= code <= 0xE01E:
+                elif 0xE0100 <= code <= 0xE01EF:
                     # variation selectors supplement
                     length += 1
                 elif 0x180B <= code <= 0x180D:
@@ -188,6 +211,7 @@ class StsDict(OrderedDict):
         if sort: iterator = sorted(iterator)
         for key, values in iterator:
             f.write(f'{key}\t{" ".join(values)}\n')
+        if f is not sys.stdout: f.close()
 
     def loadjson(self, file):
         """Load from a JSON file.
@@ -210,6 +234,7 @@ class StsDict(OrderedDict):
         """
         f = open(file, "w", encoding="UTF-8", newline="") if file else sys.stdout
         json.dump(self, f, ensure_ascii=False, indent=indent, sort_keys=sort)
+        if f is not sys.stdout: f.close()
 
     def print(self, sort=False):
         """Print key-values pairs.
@@ -296,28 +321,27 @@ class StsDict(OrderedDict):
         """Prefix self with stsdict.
 
         Convert keys of self using swapped stsdict.
-        (Enumerate all matches.)
+        (Enumerate all potential matches.)
 
         Returns:
             a new object with the same class.
 
         e.g.
-        self:
+        table:
             註冊表 => 登錄檔
         stsdict:
             注 => 注 註
         swapped stsdict:
             注 => 注
             註 => 注
-        self._join_prefix(stsdict):
-            註冊表 => 登錄檔
+        table._join_prefix(stsdict):
             注冊表 => 登錄檔
+            註冊表 => 登錄檔
         """
         dict_ = self.__class__()
         converter = stsdict.swap()
         for key, values in self.iter():
-            dict_.add(key, values)
-            for newkey in converter.apply_enum(key, no_short=False):
+            for newkey in converter.apply_enum(key, include_short=True, include_self=True):
                 dict_.add(newkey, values)
         return dict_
 
@@ -325,23 +349,24 @@ class StsDict(OrderedDict):
         """Postfix self with stsdict.
 
         Convert values of self using stsdict.
-        (Enumerate all maximal matches.)
+        (Enumerate maximal matches.)
 
         Returns:
             a new object with the same class.
 
         e.g.
-        self:
+        table:
             因为 => 因爲
         stsdict:
             爲 => 為
-        self._join_postfix(stsdict):
+        table._join_postfix(stsdict):
             因为 => 因為
+            爲 => 為
         """
         dict_ = self.__class__()
         for key, values in self.iter():
             for value in values:
-                dict_.add(key, stsdict.apply_enum(value, no_short=True))
+                dict_.add(key, stsdict.apply_enum(value))
         dict_.add_dict(stsdict)
         return dict_
 
@@ -358,6 +383,7 @@ class StsDict(OrderedDict):
                 parts if isinstance(parts, list) else
                 list(parts))
         i = max(len(Unicode.split(key)) for key in self)
+        i = min(i, len(parts) - pos)
         while i >= 1:
             end = pos + i
             current = "".join(parts[pos:end])
@@ -390,27 +416,32 @@ class StsDict(OrderedDict):
                 yield parts[i]
                 i += 1
 
-    def apply_enum(self, parts, no_short=False):
+    def apply_enum(self, parts, include_short=False, include_self=False):
         """Enumerate all possible conversions of text.
 
         Args:
             parts: a string or iterable parts to be converted.
-            no_short: output conversions that are maximally-matched
+            include_short: include non-maximal-match conversions
+            include_self: include source for every match
 
         Returns:
             list: a list of possible conversions.
 
         e.g.
         table:
-            干 => 干 乾
-            于 => 于 於
-            干于 => 甘芋 乾芋
+            钟 => 鍾 鐘
+            药 => 藥 葯
+            用药 => 用藥
         text:
-            '干于'
-        table.apply_enum(text, no_short=False):
-            ['甘芋', '乾芋', '干于', '干於', '乾于', '乾於']
-        table.apply_enum(text, no_short=True):
-            ['甘芋', '乾芋']
+            '看钟用药'
+        table.apply_enum(text, include_short=False, include_self=False):
+            ['看鍾用藥', '看鐘用藥']
+        table.apply_enum(text, include_short=True, include_self=False):
+            ['看鍾用藥', '看鐘用藥', '看鍾用葯', '看鐘用葯']
+        table.apply_enum(text, include_short=False, include_self=True):
+            ['看鍾用藥', '看鍾用药', '看鐘用藥', '看鐘用药', '看钟用藥', '看钟用药']
+        table.apply_enum(text, include_short=True, include_self=True):
+            ['看鍾用藥', '看鍾用药', '看鐘用藥', '看鐘用药', '看钟用藥', '看钟用药', '看鍾用葯', '看鐘用葯', '看钟用葯']
         """
         if isinstance(parts, str):
             text = parts
@@ -426,7 +457,7 @@ class StsDict(OrderedDict):
         while len(queue):
             (parts, matched, nextindex) = data = queue.pop(0)
             if nextindex < len(parts):
-                queue += self._apply_enum_sub(data, no_short=no_short)
+                queue += self._apply_enum_sub(data, include_short=include_short, include_self=include_self)
             elif matched > 0:
                 results["".join(parts)] = True
 
@@ -437,7 +468,7 @@ class StsDict(OrderedDict):
 
         return results
 
-    def _apply_enum_sub(self, data, no_short=False):
+    def _apply_enum_sub(self, data, include_short=False, include_self=False):
         """Helper function of apply_enum
 
         Args:
@@ -448,16 +479,49 @@ class StsDict(OrderedDict):
         """
         (parts, matched, index) = data
         results = []
+        has_atomic_match = False
         if index < len(parts):
             match = self.match(parts, index)
             if match is not None:
+                if match.end - index == 1:
+                    has_atomic_match = True
                 for value in match.conv.values:
                     result = parts[:index] + [value] + parts[match.end:]
-                    results.append((result, matched + 1, match.end))
+                    results.append((result, matched + 1, index + 1))
+                if include_self and match.conv.key not in match.conv.values:
+                    result = parts[:index] + [match.conv.key] + parts[match.end:]
+                    results.append((result, matched + 1, index + 1))
 
-                if no_short:
+                if not include_short:
                     return results
 
+                # add shorter matches
+                for i in range(match.end - match.start - 1, 0, -1):
+                    match = self.match(parts[:index + i], index)
+                    if match is not None:
+                        if match.end - index == 1:
+                            has_atomic_match = True
+                        for value in match.conv.values:
+                            result = parts[:index] + [value] + parts[match.end:]
+                            results.append((result, matched + 1, index + 1))
+                        if include_self and match.conv.key not in match.conv.values:
+                            result = parts[:index] + [match.conv.key] + parts[match.end:]
+                            results.append((result, matched + 1, index + 1))
+
+                # add atomic match (length = 1) case
+                #
+                # e.g.
+                # table: 采信 => 採信, 信息 => 訊息
+                # text: ["采", "信", "息"]
+                #
+                # We get a match ["采", "信", "息"] but no atomic match available.
+                #                 ^^^^^^^^^^
+                #
+                # Add ["采", "信", "息"] so that "采訊息" is not missed.
+                #             ^
+                if not has_atomic_match:
+                    results.append((parts, matched, index + 1))
+
         if len(results) == 0:
             results.append((parts, matched, index + 1))
 
@@ -498,6 +562,7 @@ class Table(StsDict):
                 list(parts))
         if parts[pos][0] in self.key_headchars:
             i = self.key_maxlen
+            i = min(i, len(parts) - pos)
             while i >= 1:
                 end = pos + i
                 current = "".join(parts[pos:end])
@@ -586,7 +651,7 @@ class Trie(StsDict):
 class StsListMaker():
     """A class for compling a dictionary.
     """
-    def make(self, config, format=None, output_dir=None, quiet=False):
+    def make(self, config, output_dir=None, quiet=False):
         """Compile a dictionary according to config.
 
         Load dictionaries specified in config and generate a new dictionary.
@@ -594,19 +659,13 @@ class StsListMaker():
         scheme of config (.json):
         {
             "name": "...",
-            "format": "...",
-            "dicts": [
-                [file1-1, file1-2, ...]
-                [file2-1, file2-2, file2-3, ...]
-                [file3-1, file3-2, file3-3, ...]
-            ]
+            "dicts": [{
+                "file": "...",  // relative to config directory
+                "type": "...", // list, jlist, tlist
+                "mode": "...", // load, swap, join
+                "src": ["...", ...] // list of .txt or .list files
+            }, ...]
         }
-
-        file#-# is a dictionary file (.txt) or a directory with dictionary
-        files. It can be an absolute path, relative to the config file, or
-        relative to dictionary directory. If a directory is specified, 
-        containing *.txt files will be merged in the order of their UTF-8
-        filename.
         """
         def get_config_file(config):
             """Calculate the path of a config file.
@@ -637,7 +696,7 @@ class StsListMaker():
 
             return config
 
-        def get_stsdict_path(stsdict, config_dir):
+        def get_stsdict_path(stsdict):
             """Calculate the path of a dictionary file.
 
             1. Use it if it's an absolute path.
@@ -661,7 +720,7 @@ class StsListMaker():
         # Load config
         config_file = get_config_file(config)
         if not os.path.isfile(config_file):
-            sys.exit('[error] specified config "' + config + '" is not a file')
+            sys.exit(f'[error] specified config "{config}" is not a file')
 
         config_dir = os.path.abspath(os.path.dirname(config_file))
         dictionary_dir = os.path.abspath(os.path.join(__file__, '..', 'data', 'dictionary'))
@@ -670,77 +729,54 @@ class StsListMaker():
             config = json.load(f)
             f.close()
 
-        format = format or config.get('format') or 'tlist'
-        fileslist = config['dicts']
-
-        # Merge referred dictionaries (generate only updated ones)
-        if not quiet: print('merging conversion lists...')
-
-        def check_update(output, files):
+        def check_update(output, filegroups):
             if not os.path.isfile(output):
                 return True
 
-            for file in files:
-                if os.path.getmtime(file) > os.path.getmtime(output):
-                    return True
-
-            return False
-
-        for paths in fileslist:
-            for i, path in enumerate(paths):
-                path = get_stsdict_path(path, config_dir)
-
-                if os.path.isdir(path):
-                    merged = paths[i] = path + '.list'
-                    files = sorted(glob(path + '/*.txt'))
-
-                    # Check for updates and generate only if yes.
-                    if check_update(merged, files):
-                        if not quiet: print(f'  merge: {merged} <= {", ".join(files)}')
-                        Table().load(*files).dump(merged)
-                    else:
-                        if not quiet: print(f'  skip merge (up-to-date): {merged}')
-                elif os.path.isfile(path):
-                    paths[i] = path
-
-        # Compile final dictionary.
-        if not quiet: print('making list files...')
-
-        def check_update(output, fileslist):
-            if not os.path.isfile(output):
-                return True
-
-            for files in fileslist:
+            for files in filegroups:
+                if type(files) is str:
+                    files = [files]
                 for file in files:
                     if os.path.getmtime(file) > os.path.getmtime(output):
                         return True
 
             return False
 
-        output_dir = output_dir or config_dir
-        output_base, _ = os.path.splitext(os.path.basename(config_file))
-        output_ext = '.' + format
-        output_file = os.path.abspath(os.path.join(output_dir, output_base + output_ext))
+        for dict_ in config['dicts']:
+            dest = os.path.join(output_dir or config_dir, dict_['file'])
+            format = dict_['format']
+            mode = dict_['mode']
+            files = [get_stsdict_path(f) if type(f) is str
+                else [get_stsdict_path(i) for i in f]
+                for f in dict_['src']]
 
-        # Print used tables.
-        for i, files in enumerate(fileslist):
-            if not quiet: print(f'  table {i+1}: {", ".join(files)}')
+            if not check_update(dest, files):
+                if not quiet: print(f'skip making (up-to-date): {dest}')
+                continue
+
+            if not quiet: print(f'making: {dest}')
+
+            if mode == 'load':
+                table = Table().load(*files)
+            elif mode == 'swap':
+                table = Table().load(*files).swap()
+            elif mode == 'join':
+                table = Table().load_fileslist(files)
+            else:
+                raise ValueError(f'Specified mode "{mode}" is not supported.')
+
+            os.makedirs(os.path.dirname(dest), exist_ok=True)
 
-        # Check for updates and generate only if yes.
-        if check_update(output_file, fileslist):
-            if not quiet: print('  output list: ' + output_file)
-            os.makedirs(output_dir, exist_ok=True)
             if format == 'list':
-                Table().load_fileslist(fileslist).dump(output_file)
+                table.dump(dest)
             elif format == 'jlist':
-                Table().load_fileslist(fileslist).dumpjson(output_file)
-            else: # default: tlist
-                # Trie().load_fileslist(fileslist) is 2x slower
-                Trie().add_dict(Table().load_fileslist(fileslist)).dumpjson(output_file)
-        else:
-            if not quiet: print(f'  skip output list (up-to-date): {output_file}')
+                table.dumpjson(dest)
+            elif format == 'tlist':
+                Trie().add_dict(table).dumpjson(dest)
+            else:
+                raise ValueError(f'Specified format "{format}" is not supported.')
 
-        return output_file
+        return dest
 
 class StsConverter():
     """Convert a text using a listfile.
@@ -759,7 +795,46 @@ class StsConverter():
         """Convert a text.
 
         Returns:
-            a generator of converted text content.
+            a generator of converted data.
+        """
+        try:
+            regex = re.compile(self.options['exclude'])
+        except TypeError:
+            conversion = self.table.apply(text)
+        except re.error as ex:
+            raise ValueError(f'regex syntax error for --exclude: {ex}')
+        else:
+            def convert_with_regex(text, regex):
+                index = 0
+                for m in regex.finditer(text):
+                    start, end = m.start(0), m.end(0)
+
+                    t = text[index:start]
+                    if t:
+                        yield from self.table.apply(t)
+
+                    t = text[start:end]
+                    if t:
+                        try:
+                            yield m.group('return')
+                        except IndexError:
+                            yield t
+
+                    index = end
+
+                t = text[index:]
+                if t:
+                    yield from self.table.apply(t)
+
+            conversion = convert_with_regex(text, regex)
+
+        yield from conversion
+
+    def convert_text(self, text):
+        """Convert a text.
+
+        Returns:
+            converted text content.
         """
         def parts_to_text(parts):
             for part in parts:
@@ -807,40 +882,10 @@ class StsConverter():
                     part = f'''<span tabindex="0" class="{' '.join(classes)}">{content}</span>'''
                     yield part
 
-
-        try:
-            regex = re.compile(self.options['exclude'])
-        except TypeError:
-            conversion = self.table.apply(text)
-        except re.error as ex:
-            raise ValueError(f'regex syntax error for --exclude: {ex}')
-        else:
-            def convert_with_regex(text, regex):
-                index = 0
-                for m in regex.finditer(text):
-                    start, end = m.start(0), m.end(0)
-
-                    t = text[index:start]
-                    if t:
-                        yield from self.table.apply(t)
-
-                    t = text[start:end]
-                    if t:
-                        try:
-                            yield m.group('return')
-                        except IndexError:
-                            yield t
-
-                    index = end
-
-                t = text[index:]
-                if t:
-                    yield from self.table.apply(t)
-
-            conversion = convert_with_regex(text, regex)
+        conversion = self.convert(text)
 
         if self.options['format'] == 'htmlpage':
-            yield '''<!DOCTYPE html>
+            return '''<!DOCTYPE html>
 <html>
 <head>
 <meta charset="UTF-8">
@@ -854,17 +899,15 @@ ins, del { text-decoration: none; }
 </style>
 </head>
 <body>
-<pre contenteditable="true">'''
-            yield from parts_to_html(conversion)
-            yield '''</pre>
+<pre contenteditable="true">''' + ''.join(parts_to_html(conversion)) + '''</pre>
 </body>
 </html>'''
         elif self.options['format'] == 'html':
-            yield from parts_to_html(conversion)
+            return ''.join(parts_to_html(conversion))
         elif self.options['format'] == 'json':
-            yield json.dumps(list(conversion), ensure_ascii=False)
+            return json.dumps(list(conversion), ensure_ascii=False)
         else:  # default: format = 'txt'
-            yield from parts_to_text(conversion)
+            return ''.join(parts_to_text(conversion))
 
     def convert_file(self, input=None, output=None):
         """Convert input and write to output.
@@ -877,16 +920,16 @@ ins, del { text-decoration: none; }
         text = f.read()
         f.close()
 
-        conversion = self.convert(text)
+        conversion = self.convert_text(text)
 
         f = open(output, "w", encoding="UTF-8", newline="") if output else sys.stdout
-        for part in conversion:
-            f.write(part)
+        f.write(conversion)
         f.close()
 
     default_options={
         'format': 'txt',
         'mark': False,
+        'exclude': None,
         }
 
 def main():
@@ -926,13 +969,12 @@ def main():
     def make(args):
         """Compile conversion dictionary(s).
         """
-        configs = args['config']
-        format = args['format']
+        configs = ['_default'] + args['config']
         dir = args['dir']
         quiet = args['quiet']
 
         for config in configs:
-            StsListMaker().make(config, format=format, output_dir=dir, quiet=quiet)
+            StsListMaker().make(config, output_dir=dir, quiet=quiet)
 
     def convert(args):
         """Convert a file using the given config.
@@ -946,6 +988,7 @@ def main():
             'exclude': args['exclude'],
             }
 
+        StsListMaker().make('_default', quiet=True)
         stsdict = StsListMaker().make(config, quiet=True)
         converter = StsConverter(stsdict, options)
         converter.convert_file(input, output)
@@ -1010,13 +1053,10 @@ def main():
     # subcommand: make
     parser_make = subparsers.add_parser('make',
         help=make.__doc__, description=make.__doc__)
-    parser_make.add_argument('config', nargs='+',
-        help="""the config file(s) to compile""")
-    parser_make.add_argument('--format', '-f', default=None,
-        choices=['list', 'jlist', 'tlist'], metavar='FORMAT',
-        help="""output format (list|jlist|tlist) (default: tlist)""")
+    parser_make.add_argument('config', nargs='*',
+        help="""the config file(s) to compile (default: built-in compile)""")
     parser_make.add_argument('--dir', '-d', default=None,
-        help="""the directory to save the output (default: as config directory)""")
+        help="""the directory to save the output (default: config directory)""")
     parser_make.add_argument('--quiet', '-q', default=False, action='store_true',
         help="""do not show process information""")
 
